(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Conditional_Rewriting
imports
  Trs
  Matching
  Reduction_Pair
begin

type_synonym ('f, 'v) condition = "('f, 'v) term \<times> ('f, 'v) term"
type_synonym ('f, 'v) crule = "('f, 'v) rule \<times> ('f, 'v) condition list"
type_synonym ('f, 'v) ctrs = "('f, 'v) crule set"

fun cond_step_n :: "('f, 'v) ctrs \<Rightarrow> nat \<Rightarrow> ('f, 'v) trs"
  where "cond_step_n R 0 = {}"
    |   "cond_step_n R (Suc n) = {(C\<langle>l \<cdot> \<sigma>\<rangle>,C\<langle>r \<cdot> \<sigma>\<rangle>) | C l r \<sigma> cs. ((l, r), cs) \<in> R \<and> 
            (\<forall> (si, ti) \<in> set cs. (si \<cdot> \<sigma>, ti \<cdot> \<sigma>) \<in> (cond_step_n R n)^*)}"

lemma cond_step_nI[intro]: "((l,r),cs) \<in> R \<Longrightarrow>
  (\<And> si ti.  (si,ti) \<in> set cs \<Longrightarrow> (si \<cdot> \<sigma>, ti \<cdot> \<sigma>) \<in> (cond_step_n R n)^*) \<Longrightarrow> s = C \<langle> l \<cdot> \<sigma> \<rangle> 
  \<Longrightarrow> t = C \<langle> r \<cdot> \<sigma> \<rangle> \<Longrightarrow> (s,t) \<in> cond_step_n R (Suc n)"
  unfolding cond_step_n.simps by blast

lemma cond_step_nE[elim]: "(s, t) \<in> cond_step_n R (Suc n) \<Longrightarrow>
  (\<And>l r cs \<sigma> C. \<lbrakk> ((l,r),cs) \<in> R;
  (\<And> si ti.  (si,ti) \<in> set cs \<Longrightarrow> (si \<cdot> \<sigma>, ti \<cdot> \<sigma>) \<in> (cond_step_n R n)^*); s = C \<langle> l \<cdot> \<sigma> \<rangle>;
  t = C \<langle> r \<cdot> \<sigma> \<rangle>\<rbrakk> \<Longrightarrow> P) \<Longrightarrow> P"
  unfolding cond_step_n.simps by blast

lemma cond_step_n_ctxt_subst: assumes step: "(s,t) \<in> cond_step_n R n"
  shows "(D \<langle>s \<cdot> \<sigma> \<rangle>,D \<langle>t \<cdot> \<sigma> \<rangle>) \<in> cond_step_n R n"
  using step
proof (induct n arbitrary: s t D)
  case 0
  thus ?case by simp
next
  case (Suc n s t D)
  from Suc(2) obtain l r cs C \<mu> where lr: "((l,r),cs) \<in> R" and
    cs: "\<And> si ti.  (si,ti) \<in> set cs \<Longrightarrow> (si \<cdot> \<mu>, ti \<cdot> \<mu>) \<in> (cond_step_n R n)^*"
    and s: "s = C \<langle> l \<cdot> \<mu> \<rangle>" and t: "t = C \<langle> r \<cdot> \<mu> \<rangle>" by blast
  show ?case unfolding s t
  proof (rule cond_step_nI[OF lr, of "\<mu> \<circ>s \<sigma>" _ _ "D \<circ>c (C \<cdot>c \<sigma>)"])
    fix si ti 
    {
      fix s t
      assume "(s,t) \<in> (cond_step_n R n)^*"
      hence "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> (cond_step_n R n)^*"
      proof (induct)
        case base show ?case by simp
      next
        case (step t u)
        from step(3) Suc(1)[OF step(2), of Hole] show ?case by auto
      qed
    } note main = this
    assume "(si, ti) \<in> set cs" 
    from main[OF cs[OF this]]
    show "(si \<cdot> \<mu> \<circ>s \<sigma>, ti \<cdot> \<mu> \<circ>s \<sigma>) \<in> (cond_step_n R n)^*" by simp
  qed auto
qed

definition cond_step :: "('f,'v)ctrs \<Rightarrow> ('f,'v)trs" 
  where "cond_step R \<equiv> \<Union>n. cond_step_n R n"

definition quasi_reductive :: "('f,'v)ctrs \<Rightarrow> bool"
  where "quasi_reductive R \<equiv> \<exists> S NS. redpair_order S NS \<and> ctxt_closed S \<and> (\<forall> l r cs (\<sigma> :: ('f,'v)subst). ((l,r),cs) \<in> R \<longrightarrow> 
         (\<forall> i. i < length cs 
         \<longrightarrow> (\<forall> j. j < i \<longrightarrow> (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> S^=) 
         \<longrightarrow> (l \<cdot> \<sigma>, fst (cs ! i) \<cdot> \<sigma>) \<in> (S \<union> {\<rhd>})^+) \<and> 
         ((\<forall> j. j < length cs \<longrightarrow> (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> S^=)
          \<longrightarrow> (l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> S)
        )"

lemma quasi_reductive_SN: assumes qr: "quasi_reductive (R :: ('f,'v)ctrs)"
  shows "SN (cond_step R)"
proof -
  from qr[unfolded quasi_reductive_def]
  obtain S NS where rp: "redpair_order S NS"
  and mono: "ctxt_closed S"
  and qr: "\<And> l r cs (\<sigma> :: ('f,'v)subst).
             ((l, r), cs) \<in> R \<Longrightarrow>
             (\<forall>j<length cs. (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> S^=) \<Longrightarrow> (l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> S" by blast
  interpret redpair_order S NS by fact
  show ?thesis
  proof (rule SN_subset[OF SN])
    show "cond_step R \<subseteq> S"
    proof
      fix s t
      assume "(s,t) \<in> cond_step R"
      then obtain n where "(s,t) \<in> cond_step_n R n" unfolding cond_step_def by auto
      thus "(s,t) \<in> S"
      proof (induct n arbitrary: s t)
        case 0 thus ?case by simp
      next
        case (Suc n s t)
        from Suc(2)
        obtain l r cs C \<sigma> where lr: "((l,r),cs) \<in> R" and
          cs: "\<And> si ti.  (si,ti) \<in> set cs \<Longrightarrow> (si \<cdot> \<sigma>, ti \<cdot> \<sigma>) \<in> (cond_step_n R n)^*"
          and s: "s = C \<langle> l \<cdot> \<sigma> \<rangle>" and t: "t = C \<langle> r \<cdot> \<sigma> \<rangle>" by blast
        note qr = qr[OF lr, of \<sigma>]
        have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> S"
        proof (rule qr[rule_format])
          fix j
          assume j: "j < length cs"
          obtain sj tj where csj: "cs ! j = (sj,tj)" by force
          from j have "(sj,tj) \<in> set cs" unfolding csj[symmetric] set_conv_nth by auto
          from cs[OF this] have steps: "(sj \<cdot> \<sigma>, tj \<cdot> \<sigma>) \<in> (cond_step_n R n)^*" .
          {
            fix s t
            assume "(s,t) \<in> (cond_step_n R n)^*"
            hence "(s,t) \<in> S^="
            proof (induct)
              case base thus ?case by simp
            next
              case (step t u)
              from step(3) Suc(1)[OF step(2)] have "(s,u) \<in> S^= O S" by auto
              thus "(s,u) \<in> S^=" using trans_S_point by blast
            qed
          }
          from this[OF steps]
          show "(fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> S^=" unfolding csj by simp
        qed
        from ctxt_closedE[OF mono this, of C]
        show ?case unfolding s t .
      qed
    qed
  qed
qed
     
locale unraveling =  
  fixes R :: "('f,'v)ctrs"
  and U :: "('f,'v)crule \<Rightarrow> nat \<Rightarrow> ('f,'v)ctxt"
  (* e.g., U ((l,r),s1 = t1, ... sn = tn) i = U_i(\<box>,vars(l) \<union> vars(t1) ... \<union> vars(t i-1) *)
begin

fun lhs_n :: "('f,'v)crule \<Rightarrow> nat \<Rightarrow> ('f,'v)term"
  where "lhs_n (lr,cs) 0 = fst lr"
      | "lhs_n (lr,cs) (Suc n) = (U (lr,cs) n) \<langle>snd (cs ! n)\<rangle>"

fun rhs_n :: "('f,'v)crule \<Rightarrow> nat \<Rightarrow> ('f,'v)term"
  where "rhs_n (lr,cs) n = (if n < length cs then 
         (U (lr,cs) n)\<langle>fst (cs ! n)\<rangle>
         else snd lr)"

definition rules :: "('f,'v)crule \<Rightarrow> ('f,'v)trs"
  where "rules crule \<equiv> (\<lambda> n. (lhs_n crule n, rhs_n crule n)) ` {n . n \<le> length (snd crule)}"

definition UR :: "('f, 'v) trs" where
  "UR \<equiv> \<Union>(rules ` R)"

lemma UR_simulation_main: assumes lr: "((l, r), cs) \<in> R"
  and i: "i \<le> length cs"
  and cs: "\<And> j. j < i \<Longrightarrow> (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> (rstep UR)^*"
  shows "(l \<cdot> \<sigma>, rhs_n ((l,r),cs) i \<cdot> \<sigma>) \<in> (rstep UR)^+"
proof -
  let ?lr = "((l,r),cs)"
  let ?n = "length cs"
  from lr have lr: "rules ?lr \<subseteq> UR" unfolding UR_def by auto
  hence lr: "\<And> i. i \<le> ?n \<Longrightarrow> (lhs_n ?lr i, rhs_n ?lr i) \<in> UR" unfolding rules_def
    by auto
  show ?thesis
  proof (rule rtrancl_into_trancl2)
    show "(l \<cdot> \<sigma>,rhs_n ?lr 0 \<cdot> \<sigma>) \<in> rstep UR"
      by (rule rstepI[OF lr[of 0], of _ Hole \<sigma>], auto)
    from i cs
    show "(rhs_n ?lr 0 \<cdot> \<sigma>, rhs_n ?lr i \<cdot> \<sigma>) \<in> (rstep UR)^*"
    proof (induct i)
      case 0 thus ?case by simp
    next
      case (Suc i)
      hence i: "i < ?n" and i': "i \<le> ?n" by auto
      let ?f = "U ?lr i"
      let ?U = "\<lambda> t. ?f \<langle> t \<rangle>"
      from Suc(1)[OF i' Suc(3)]
      have "(rhs_n ?lr 0 \<cdot> \<sigma>, rhs_n ?lr i \<cdot> \<sigma>) \<in> (rstep UR)^*" by simp
      also have "rhs_n ?lr i = ?U (fst (cs ! i))"
        using i by simp
      also have "(?U (fst (cs ! i)) \<cdot> \<sigma>, ?U (snd (cs ! i)) \<cdot> \<sigma>) \<in> (rstep UR)^*" 
      proof -
        obtain si ti where csi: "cs ! i = (si,ti)" by force
        from Suc(3)[of i, unfolded csi] have "(si \<cdot> \<sigma>, ti \<cdot> \<sigma>) \<in> (rstep UR)^*" by simp
        from ctxt_closedE[OF ctxt_closed_rtrancl[OF ctxt_closed_rstep] this]
        have "\<And> C. (C\<langle>si \<cdot> \<sigma>\<rangle>, C\<langle>ti \<cdot> \<sigma>\<rangle>) \<in> (rstep UR)^*" .
        from this[of "?f \<cdot>c \<sigma>"]
        show ?thesis unfolding csi by simp
      qed
      also have "?U (snd (cs ! i)) = lhs_n ?lr (Suc i)" 
        unfolding lhs_n.simps by simp
      also have "(lhs_n ?lr (Suc i) \<cdot> \<sigma>, rhs_n ?lr (Suc i) \<cdot> \<sigma>) \<in> (rstep UR)"
        by (rule rstepI[OF lr[OF Suc(2)], of _ Hole \<sigma>], auto)
      finally show ?case .
    qed
  qed
qed

lemma SN_UR_quasi_reductive: assumes SN: "SN (rstep UR)"
  shows "quasi_reductive R"
proof -
  let ?S = "(rstep UR)^+"
  let ?NS = "(rstep UR)^*"
  have id: "?S^= = ?NS" by simp
  show ?thesis unfolding quasi_reductive_def
  proof (rule exI[of _ ?S], rule exI[of _ ?NS], unfold id, intro conjI allI impI)
    show "redpair_order ?S ?NS"
    proof
      show "SN ?S" by (rule SN_imp_SN_trancl[OF SN])
      show "ctxt_closed ?NS" by auto
      show "subst_closed ?NS" by (rule subst_closed_rtrancl, auto) 
      show "subst_closed ?S" by (rule subst_closed_trancl, auto) 
    qed auto
    show "ctxt_closed ?S" by auto
  next
    fix l r cs \<sigma>
    assume lr: "((l, r), cs) \<in> R"
      and cs: "\<forall>j<length cs. (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> (rstep UR)^*"
    from UR_simulation_main[OF lr, of "length cs" \<sigma>] cs
    show "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> (rstep UR)^+" by simp
  next
    fix l r cs \<sigma> i
    assume lr: "((l, r), cs) \<in> R"
    and i: "i < length cs"
    and cs: "\<forall>j<i. (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> (rstep UR)^*"
    from UR_simulation_main[OF lr, of i \<sigma>] i cs
    have lm: "(l \<cdot> \<sigma>, rhs_n ((l, r), cs) i \<cdot> \<sigma>) \<in> (rstep UR)^+" (is "(?l,?m) \<in> _") by simp
    have mr: "?m \<unrhd> fst (cs ! i) \<cdot> \<sigma>" (is "_ \<unrhd> ?r")
      by (rule supteq_stable, insert i, auto)
    let ?R = "(rstep UR)^+ \<union> {\<rhd>}"
    from mr have "(?m,?r) \<in> ({\<rhd>})^=" by auto
    hence mr: "(?m,?r) \<in> ?R^*" by auto
    have lm: "(?l,?m) \<in> ?R^+"
      by (rule trancl_mono[OF lm], auto)
    from lm mr show lr: "(?l,?r) \<in> ?R^+" by auto
  qed
qed

lemma UR_complete: assumes step: "(s,t) \<in> cond_step R"
  shows "(s,t) \<in> (rstep UR)^+"
proof -
  from step[unfolded cond_step_def] obtain n where "(s,t) \<in> cond_step_n R n" ..
  thus ?thesis
  proof (induct n arbitrary: s t)
    case 0 thus ?case by simp
  next
    case (Suc n) note IH = this
    from IH(2) obtain l r cs C \<sigma> where lr: "((l,r),cs) \<in> R" and
    cs: "\<And> si ti.  (si,ti) \<in> set cs \<Longrightarrow> (si \<cdot> \<sigma>, ti \<cdot> \<sigma>) \<in> (cond_step_n R n)^*"
    and s: "s = C \<langle> l \<cdot> \<sigma> \<rangle>" and t: "t = C \<langle> r \<cdot> \<sigma> \<rangle>" by blast
    let ?lr = "((l,r),cs)"
    let ?n = "length cs"
    have rsig: "r \<cdot> \<sigma> = rhs_n ?lr ?n \<cdot> \<sigma>" by auto
    have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> (rstep UR)^+" unfolding rsig
    proof (rule UR_simulation_main[OF lr, of "length cs" \<sigma>])
      fix j
      assume j: "j < length cs"
      obtain sj tj where csj: "cs ! j = (sj,tj)" by force
      from j have "(sj,tj) \<in> set cs" unfolding csj[symmetric] set_conv_nth by auto
      from cs[OF this] have steps: "(sj \<cdot> \<sigma>, tj \<cdot> \<sigma>) \<in> (cond_step_n R n)^*" .
      {
        fix s t
        assume "(s,t) \<in> (cond_step_n R n)^*"
        hence "(s,t) \<in> (rstep UR)^*"
        proof (induct)
          case base thus ?case by auto
        next
          case (step t u)
          from step(3) IH(1)[OF step(2)] show ?case by auto
        qed
      }
      from this[OF steps]
      show "(fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> (rstep UR)^*" 
        unfolding csj by auto
    qed simp
    from ctxt_closedE[OF ctxt_closed_trancl[OF ctxt_closed_rstep] this, of C]
    show ?case unfolding s t .
  qed
qed

end

definition "eqs \<mu> = map (\<lambda>(x, t). (Var x, t)) \<mu>"

hide_const (open) eqs

abbreviation funs_ctrs :: "('f,'v)ctrs \<Rightarrow> 'f set" where
 "funs_ctrs R  \<equiv> \<Union>((\<lambda> \<rho>. funs_rule (fst \<rho>) \<union> funs_trs (set (snd \<rho>))) ` R)"

abbreviation vars_ctrs :: "('f,'v)ctrs \<Rightarrow> 'v set" where
 "vars_ctrs R  \<equiv> \<Union>((\<lambda> \<rho>. vars_rule (fst \<rho>) \<union> vars_trs (set (snd \<rho>))) ` R)"

context unraveling
begin

abbreviation ss :: "('f,'v) crule \<Rightarrow> ('f,'v) term set" where "ss \<rho> \<equiv> lhss (set (snd \<rho>))"
abbreviation ts :: "('f,'v) crule \<Rightarrow> ('f,'v) term set" where "ts \<rho> \<equiv> rhss (set (snd \<rho>))"
abbreviation clhs ::  "('f,'v) crule \<Rightarrow> ('f,'v) term" where "clhs \<rho> \<equiv> fst (fst \<rho>)"
abbreviation crhs ::  "('f,'v) crule \<Rightarrow> ('f,'v) term" where "crhs \<rho> \<equiv> snd (fst \<rho>)"

(* start indexing conditions from 0 *)
(* predicate for valid X variable enumeration *)
definition X_vars :: "(('f,'v)crule \<Rightarrow> nat \<Rightarrow> 'v list) \<Rightarrow> bool" where
 "X_vars X = (\<forall> \<rho> \<in> R. \<forall> i < length (snd \<rho>). (
  set (X \<rho> i) = vars_term (fst (fst \<rho>)) \<union> \<Union>(vars_term ` (rhss (set (List.take i (snd \<rho>)))))) \<and> 
   distinct (X \<rho> i))"

(* predicate for valid Y variable enumeration *)
definition Y_vars :: "(('f,'v)crule \<Rightarrow> nat \<Rightarrow> 'v list) \<Rightarrow> bool" where
 "Y_vars Y = (\<forall> \<rho> \<in> R. \<forall> i < length (snd \<rho>). (
  set (Y \<rho> i) = vars_term (snd (fst \<rho>)) \<union> vars_term (snd (snd \<rho>!i)) \<union> 
    (vars_trs (set (List.drop i (snd \<rho>)) ))) \<and> distinct (Y \<rho> i))"

definition Z_vars :: "(('f,'v)crule \<Rightarrow> nat \<Rightarrow> 'v list) \<Rightarrow> (('f,'v)crule \<Rightarrow> nat \<Rightarrow> 'v list) \<Rightarrow> (('f,'v)crule \<Rightarrow> nat \<Rightarrow> 'v list) \<Rightarrow> bool" where
 "Z_vars Z X Y = (X_vars X \<and> Y_vars Y \<and> (\<forall> \<rho> \<in> R. \<forall> i < length (snd \<rho>).
   set (Z \<rho> i) =  set (X \<rho> i) \<inter> set (Y \<rho> i)))"

abbreviation ultra :: "(('f,'v) trs \<Rightarrow> bool) \<Rightarrow> bool" where "ultra P \<equiv> P (UR)"

abbreviation ultra_rule :: "(('f,'v) trs \<Rightarrow> bool) \<Rightarrow> ('f,'v) crule \<Rightarrow> bool" 
 where "ultra_rule P \<rho> \<equiv> P (rules \<rho>)"

 
 
end

(* locale demanding conditions for U_opt *)
locale unraveling_opt = unraveling R U 
  for R :: "('f,'v) ctrs" and U :: "('f,'v) crule \<Rightarrow> nat => ('f,'v) ctxt" +
  fixes X :: "('f,'v)crule \<Rightarrow> nat \<Rightarrow> 'v list"
  fixes Y :: "('f,'v)crule \<Rightarrow> nat \<Rightarrow> 'v list"
  fixes Z :: "('f,'v)crule \<Rightarrow> nat \<Rightarrow> 'v list"
  assumes U_cond:"\<forall> \<rho> \<in> R. \<forall>n < length (snd \<rho>). \<exists> f. 
   (U \<rho> n = (More f Nil Hole (map Var (Z \<rho> n))) \<and> f \<notin> funs_ctrs R \<and>
   \<not>(\<exists> \<rho>' n' b c a. \<rho>' \<in> R \<and> n' < length (snd \<rho>') \<and> (\<rho> \<noteq> \<rho>' \<or> n \<noteq> n') \<and> 
     U \<rho>' n' = More f b c a))"
  and Z:"Z_vars Z X Y"
  and dctrs:"\<forall> \<rho> \<in> R. \<forall> i < length (snd \<rho>). 
   vars_term (fst ((snd \<rho>)!i)) \<subseteq> (vars_term (clhs \<rho>) \<union> \<Union> ((\<lambda> j. vars_term (snd ((snd \<rho>)!j))) ` {j. j < i}))"
  and type3:"\<forall> \<rho> \<in> R. vars_term (snd (fst \<rho>)) \<subseteq> (vars_term (clhs \<rho>) \<union> \<Union> ((\<lambda> j. vars_term (snd ((snd \<rho>)!i))) ` {j. j < length (snd \<rho>)}))"
begin


(* some properties of X, Y, Z variables *)

lemma X_vars_mono :
 assumes inR:"\<rho> \<in> R" and il:"i < length (snd \<rho>)" and ji:"j \<le> i"
 shows "set (X \<rho> j) \<subseteq> set (X \<rho> i)"
proof(cases "j = i", simp)
 case False
 with ji have ji:"j < i" by auto
 let ?v = "\<lambda> m. \<Union>(vars_term ` rhss (set (take m (snd \<rho>))))"
 from Z[unfolded Z_vars_def X_vars_def] il inR have
  si:"set (X \<rho> i) = vars_term (clhs \<rho>) \<union> ?v i" by auto
 from Z[unfolded Z_vars_def X_vars_def] xt1(10)[OF il ji] inR have
  sj:"set (X \<rho> j) = vars_term (clhs \<rho>) \<union> ?v j" by meson
 have "?v j \<subseteq> ?v i" by (rule Union_mono, rule image_mono, rule image_mono, 
  simp add:set_take_subset_set_take[OF less_imp_le[OF ji]]) 
 from Un_mono[OF subset_refl this] show ?thesis unfolding si sj by auto
qed

lemma l_vars_X_vars :
 assumes inR:"\<rho> \<in> R" and il:"i < length (snd \<rho>)" shows "vars_term (clhs \<rho>) \<subseteq> set (X \<rho> i)" 
using Z[unfolded Z_vars_def X_vars_def] il inR by blast

lemma Y_vars_mono :
 assumes inR:"\<rho> \<in> R" and il:"i < length (snd \<rho>)" and ji:"j \<le> i"
 shows "set (Y \<rho> j) \<supseteq> set (Y \<rho> i)"
proof(cases "j = i", simp)
 case False
 with ji have ji:"j < i" by auto
 from ji obtain k where i:"i = j + k " 
by (metis add_Suc_right less_iff_Suc_add)
 let ?v = "\<lambda> m. (vars_trs (set (List.drop m (snd \<rho>)) ))"
 let ?r = "vars_term (snd (fst \<rho>))"
 let ?ti = "\<lambda> k. vars_term (snd (snd \<rho>!k))"
 from Z[unfolded Z_vars_def Y_vars_def] il inR have
  si:"set (Y \<rho> i) = ?r \<union> ?ti i \<union> ?v i" by auto
 from Z[unfolded Z_vars_def Y_vars_def] xt1(10)[OF il ji] inR have
  sj:"set (Y \<rho> j) = ?r \<union> ?ti j \<union> ?v j" by meson 
 from set_drop_subset_set_drop[OF less_imp_le[OF ji]] 
  have 1:"?v i \<subseteq> ?v j" unfolding vars_trs_def Union_image_eq[symmetric] by fast
 from il[unfolded i] have "k < length (drop j (snd \<rho>))" unfolding length_drop by linarith
 from nth_drop[OF less_imp_le[OF il[unfolded i]]] nth_mem[OF this] 
  have "(snd \<rho> ! i) \<in> set (drop j (snd \<rho>))" unfolding i by auto
 hence "vars_term (snd (snd \<rho> ! i)) \<subseteq> vars_trs (set (drop j (snd \<rho>)))" 
  unfolding vars_trs_def Union_image_eq[symmetric] using vars_rule_def by (metis Sup_upper Un_subset_iff imageI)
 from Un_least[OF this 1] Un_upper2 have "?ti i \<union> ?v i \<subseteq> ?ti j \<union> ?v j" by auto
 from Un_mono[OF subset_refl this] show ?thesis unfolding si sj by auto
qed

lemma Z_imp_X :
 assumes inR:"\<rho> \<in> R" and il:"i < length (snd \<rho>)" and z:"x \<in> set (Z \<rho> i)" 
 shows "x \<in> set (X \<rho> i)"
 using conjunct2[OF conjunct2[OF Z[unfolded Z_vars_def]], rule_format, OF inR il] z by blast

lemma Z_imp_Y :
 assumes inR:"\<rho> \<in> R" and il:"i < length (snd \<rho>)" and z:"x \<in> set (Z \<rho> i)" 
 shows "x \<in> set (Y \<rho> i)"
 using conjunct2[OF conjunct2[OF Z[unfolded Z_vars_def]], rule_format, OF inR il] z by blast

lemma X_Y_imp_Z :
 assumes inR:"\<rho> \<in> R" and il:"i < length (snd \<rho>)" and x:"x \<in> set (X \<rho> i)" and y:"x \<in> set (Y \<rho> i)"
 shows "x \<in> set (Z \<rho> i)"
 using  conjunct2[OF conjunct2[OF Z[unfolded Z_vars_def]], rule_format, OF inR il] x y by blast

lemma s_imp_Y :
 assumes inR:"\<rho> \<in> R" and il:"Suc i < length (snd \<rho>)" and x:"x \<in> vars_term (fst (snd \<rho> ! (Suc i)))"
 shows "x \<in> set (Y \<rho> i)"
proof-
 from Z[unfolded Z_vars_def] have y:"Y_vars Y" by auto 
 from il have len:"1 < length (drop i (snd \<rho>))" by fastforce 
 from nth_drop[OF less_imp_le[OF il[unfolded Suc_eq_plus1]]] nth_mem[OF len]
  have aux: "(snd \<rho> ! (Suc i)) \<in> (set (drop i (snd \<rho>)))" by auto
 from x have xr:"x \<in> vars_rule (snd \<rho> ! Suc i)" unfolding vars_rule_def by blast
 have "x \<in> vars_trs (set (drop i (snd \<rho>)))" unfolding vars_trs_def Union_image_eq[symmetric] 
  by(rule UnionI[OF _ xr], insert aux, auto)
 with y[unfolded Y_vars_def, rule_format, OF inR less_trans[OF lessI il]] show ?thesis by blast
qed

lemma X_le_Z_imp_Z :
 assumes inR:"\<rho> \<in> R" and il:"i < length (snd \<rho>)" and ji:"j \<le> i" and 
 x:"x \<in> set (X \<rho> j)" and z:"x \<in> set (Z \<rho> i)"
 shows "x \<in> set (Z \<rho> j)"
proof(cases "i=j")
 case True with z show ?thesis by auto
 next
 case False
 from z Z_imp_Y[OF inR il z] ji inR il have "x \<in> set (Y \<rho> i)" by blast
 with Y_vars_mono[OF inR il, unfolded order.strict_iff_order] ji False  have "x \<in> set (Y \<rho> j)" by fast
 with x show ?thesis using conjunct2[OF conjunct2[OF Z[unfolded Z_vars_def]], rule_format, OF inR] 
  order.strict_trans1[OF ji il]  by blast
qed

lemma Z_interval :
 assumes inR:"\<rho> \<in> R" and kl:"k < length (snd \<rho>)" and ji:"j \<le> i" and ik:"i \<le> k" and 
 xj:"x \<in> set (Z \<rho> j)" and xk:"x \<in> set (Z \<rho> k)"
 shows "x \<in> set (Z \<rho> i)"
proof-
  from ji kl ik have jl:"j < length (snd \<rho>)" by auto
  from kl ik have il:"i < length (snd \<rho>)" by auto
  from X_vars_mono[OF inR il ji] Z_imp_X[OF inR jl xj] have x:"x \<in> set (X \<rho> i)" by blast
  from Y_vars_mono[OF inR kl ik] Z_imp_Y[OF inR kl xk] have "x \<in> set (Y \<rho> i)" by blast
  from X_Y_imp_Z[OF inR il x this] show ?thesis by fast
qed

lemma vars_l_in_X :
 assumes inR:"\<rho> \<in> R" and il:"0 < length (snd \<rho>)" and "x \<in> vars_term (clhs \<rho>)" shows "x \<in> set (X \<rho> 0)"
 using conjunct1[OF Z[unfolded Z_vars_def], unfolded X_vars_def, rule_format, OF inR il] assms by auto


(* Theorem 3.8(1), => in NSS12 *)
lemma L3_8a :
 assumes inR:"\<rho> \<in> R"
 shows
 "ultra_rule left_linear_trs \<rho> \<Longrightarrow> (linear_term (clhs \<rho>) \<and> (\<forall> ti \<in> ts \<rho>. linear_term ti) \<and>
      (\<forall> i. ((i < length (snd \<rho>)) \<longrightarrow> (vars_term (snd ((snd \<rho>)!i)) \<inter> set (X \<rho> i))  = {} ) ))"
proof-
 assume ll:"left_linear_trs (rules \<rho>)"
 obtain l r cs where rho:"\<rho> = ((l,r),cs)" using PairE by metis
 with ll have ll:"\<And> l' r'. (l',r') \<in> rules \<rho> \<Longrightarrow> linear_term l'" unfolding left_linear_trs_def by blast
 hence "linear_term (lhs_n \<rho> 0)"  unfolding rules_def by fast
 hence 1:"linear_term (clhs \<rho>)" unfolding rules_def using lhs_n.simps(1) prod.collapse by metis
 { fix ti
  assume "ti \<in> ts \<rho>"
  from this[unfolded image_iff] in_set_conv_nth have "\<exists> i < length (snd \<rho>) . ti = snd(snd \<rho> ! i)" by metis
  then obtain i where i:"Suc i \<le> length (snd \<rho>)" and ti:"ti = snd(snd \<rho> ! i)" by force
  have lin:"linear_term (lhs_n \<rho> (Suc i))" by(rule ll[unfolded rules_def], insert i, auto)
  from i have "i < length (snd \<rho>)" by auto
  from U_cond[rule_format, OF inR this] obtain f where "U \<rho> i = More f [] \<box> (map Var (Z \<rho> i))" by fast
  from lin[unfolded rho lhs_n.simps(2) this[unfolded rho] ctxt_apply.simps] rho have 
   "linear_term (Fun f ((snd (cs!i)) # (map Var (Z \<rho> i))))" by auto
  from this[unfolded linear_term.simps(2)] have "linear_term ti" unfolding ti rho by force
 } note 2 = this
 { fix i
   assume i:"i < length (snd \<rho>)"
   have "vars_term (snd ((snd \<rho>)!i)) \<inter> set (X \<rho> i)  = {} "
   proof(rule ccontr)
   assume ne:"vars_term (snd (snd \<rho> ! i)) \<inter> set (X \<rho> i) \<noteq> {}"
   let ?ti = "(snd (snd \<rho> ! i))"
   let ?zs = "Z \<rho> i"
   from Z[unfolded Z_vars_def Y_vars_def] have "vars_term ?ti \<subseteq> set (Y \<rho> i)" using inR i by force
   with Z[unfolded Z_vars_def] ne have iz:"vars_term ?ti \<inter> set (Z \<rho> i) \<noteq> {}" using inR i by fast
   have zs:"\<Union> (set (map vars_term (map Var ?zs))) = set ?zs" unfolding map_map vars_term_Var_id by force
   from iz have p:"\<not> (is_partition (map vars_term (?ti # (map Var ?zs))))" 
    unfolding list.simps(9) is_partition_Cons zs by blast
   let ?t = "(U \<rho> i)\<langle>snd (snd \<rho> ! i)\<rangle>"
   from U_cond[rule_format, OF inR i] obtain f where Uri:"U \<rho> i = More f [] \<box> (map Var (Z \<rho> i))" by fast
   from p have "\<not> linear_term ?t" unfolding Uri by simp
   hence l:"\<not> linear_term (lhs_n \<rho> (Suc i))" unfolding rho by auto
   have "linear_term (lhs_n \<rho> (Suc i))" by(rule ll[unfolded rules_def], insert i, auto)
   with l show False by auto
  qed 
 } note 3 = this
 from 1 2 3 show ?thesis by auto
qed

abbreviation si :: "('f,'v) crule \<Rightarrow> nat =>  ('f,'v) term" where "si \<rho> i \<equiv> fst ((snd \<rho>)!i)"
abbreviation ti :: "('f,'v) crule \<Rightarrow> nat =>  ('f,'v) term" where "ti \<rho> i \<equiv> snd ((snd \<rho>)!i)"



lemma part_rev : assumes p:"is_partition xs" shows "is_partition (rev xs)"
 unfolding is_partition_def length_rev 
proof(rule,rule)
 fix j
 assume j:"j < length xs"
 show "\<forall>i<j. rev xs ! i \<inter> rev xs ! j = {}" proof(rule,rule)
  fix i
  assume i:"i < j"
  from j have 1:"length xs  - (Suc i) < length xs" by fastforce
  from i j have 2:"length xs  - (Suc j) < length xs  - (Suc i)"
    by (metis Suc_less_eq diff_less_mono2 less_trans_Suc)
  from p[unfolded is_partition_def] 1 2 
  show "rev xs ! i \<inter> rev xs ! j  = {}" unfolding rev_nth[OF j] rev_nth[OF order.strict_trans[OF i j]] by blast
 qed
qed

lemma is_partition_rev : "is_partition (rev xs) = is_partition xs "
 using part_rev rev_rev_ident by force
    
(* to be moved somewhere else *)
lemma substs_rsteps':
  shows "(\<And>x. x \<in> vars_term t \<Longrightarrow> (\<sigma> x, \<tau> x) \<in> (rstep RR)\<^sup>*) \<Longrightarrow> (t \<cdot> \<sigma>, t \<cdot> \<tau>) \<in> (rstep RR)\<^sup>*"
proof (induct t)
  case (Var y) 
  thus ?case using assms [of y] by simp
next
  case (Fun f ts)
  have "\<And> s. s \<in> set ts \<Longrightarrow>  vars_term s \<subseteq> vars_term (Fun f ts)" by auto
  with Fun have "\<And> s. s \<in> set ts \<Longrightarrow> (s \<cdot> \<sigma>, s \<cdot> \<tau>) \<in> (rstep RR)\<^sup>*" by blast
  then have "\<forall>i<length (map (\<lambda>t. t \<cdot> \<sigma>) ts).
    (map (\<lambda>t. t \<cdot> \<sigma> ) ts ! i, map (\<lambda>t. t \<cdot> \<tau>) ts ! i) \<in> (rstep RR)\<^sup>*" by auto
  from args_rsteps_imp_rsteps [OF _ this] show ?case by simp
qed

(* Lemma A.1 in NSS12 *)
(* partition of variable sets in substitution \<sigma> (different from paper *)
definition X_part :: "('f,'v) crule \<Rightarrow> 'v set list" where
 "X_part \<rho> \<equiv> ((set (X \<rho> 0)) # (map (\<lambda>i. set (X \<rho> (Suc i)) -  (set (X \<rho> i))) [0..< (length (snd \<rho>) - 1)]) @
             [(vars_term (crhs \<rho>) \<union> (vars_term (ti \<rho> (length (snd \<rho>) - 1)))) - (set (X \<rho> (length (snd \<rho>) - 1)))])"


lemma X_part: assumes inR:"\<rho> \<in> R" 
 and l:"length (snd \<rho>) \<noteq> 0 "
 shows "is_partition (X_part \<rho>)"
proof-
 let ?k = "length (snd \<rho>)"
 from not0_implies_Suc[OF l] obtain k' where k_suc:"?k = Suc k'" by auto
 let ?xk = "(vars_term (crhs \<rho>) \<union> (vars_term (ti \<rho> k'))) - (set (X \<rho> k'))"
 let ?xs = "\<lambda> m. (set (X \<rho> 0)) # (map (\<lambda>i. set (X \<rho> (Suc i)) -  (set (X \<rho> i))) [0..<m])"
 fix j
 have key:"\<forall>j. j < ?k \<longrightarrow> (is_partition (rev (?xs j)) \<and> \<Union>set (?xs j) \<subseteq> set (X \<rho> j))" proof(rule allI, rule)
  fix j 
  assume "j < ?k"
  thus "is_partition (rev (?xs j))  \<and> \<Union>set (?xs j) \<subseteq> set (X \<rho> j)" proof(induct j)
  case 0 
   have s:"?xs 0 = [set (X \<rho> 0)]" by auto
   thus ?case unfolding s using rev.simps is_partition_Cons is_partition_Nil by force
  next
  case (Suc j)
   assume lt:"Suc j < ?k"
   with Suc have p:"is_partition (rev (?xs j))" and ss:"\<Union>set (?xs j) \<subseteq> set (X \<rho> j)" by auto
   let ?xi = "\<lambda> j. set (X \<rho> (Suc j)) -  (set (X \<rho> j))"
   have "?xs (Suc j) = (?xs j) @ [?xi j]" unfolding upt_Suc_append[OF le0] map_append by fastforce
   hence r:"rev (?xs (Suc j)) =  (?xi j) #  (rev (?xs j))" using rev_eq_Cons_iff[of "?xs (Suc j)"] by auto
   have "(?xi j) \<inter> set (X \<rho> j) = {}" by auto
   with ss have "(?xi j) \<inter> \<Union>set (rev (?xs j)) = {}" by auto
   with Suc r have p:"is_partition (rev (?xs (Suc j)))" using is_partition_Cons[of "?xi j" "rev (?xs j)"] by fastforce
   from ss X_vars_mono[OF inR lt less_imp_le[OF lessI]] 
     have "\<Union>set (?xs (Suc j)) \<subseteq> set (X \<rho> (Suc j))" 
     unfolding  upt_Suc_append[OF le0] map_append list.set(2) set_append by auto
   with p show ?case by auto
  qed
  qed
 have lt:"k' < ?k" using k_suc by auto
 from key this have p:"is_partition (rev (?xs k'))" "\<Union>set (?xs k') \<subseteq> set (X \<rho> k')" unfolding X_part_def by auto
 from p(2) have isc:"?xk \<inter> \<Union>set (rev (?xs k')) = {}" by auto
 from k_suc have kk:"?k - 1 = k'" by auto
 have aux:"rev (?xs k' @ [?xk]) = ?xk # rev (?xs k')"  using rev_eq_Cons_iff by fastforce
 with p isc have "is_partition (rev (?xs k' @ [?xk]))" unfolding aux is_partition_Cons by blast
 thus ?thesis using is_partition_rev unfolding X_part_def kk by force
qed

lemma X_part_nth :
 assumes inR:"\<rho> \<in> R" 
 and l:"length (snd \<rho>) \<noteq> 0 "
 and jk:"Suc j < length (snd \<rho>)"
 shows " (X_part \<rho>) ! (Suc j) =  set (X \<rho> (Suc j)) - set (X \<rho> j)"
proof-
 let ?k = "length (snd \<rho>)"
 let ?f = "\<lambda>i. set (X \<rho> (Suc i)) - set (X \<rho> i)"
 from jk have x:"j < ?k - 1 - 0" by auto
 hence "map ?f [0..<?k - 1] ! j = set (X \<rho> (Suc j)) - set (X \<rho> j)" 
  unfolding nth_map[of j "[0..<?k - 1]", unfolded length_upt, OF x] 
  using nth_upt by force
 with nth_map[of j "[0..<?k - 1]", unfolded length_upt, OF x] show ?thesis unfolding X_part_def nth_Cons_Suc 
   nth_append[of "map ?f [0..<?k - 1]" _ j, unfolded length_map length_upt eqTrueI[OF x] if_True] by blast
qed

lemma X_part_last :
 assumes inR:"\<rho> \<in> R" 
 and l:"length (snd \<rho>) \<noteq> 0 "
 shows " (X_part \<rho>) ! (length (snd \<rho>)) = 
  (vars_term (crhs \<rho>) \<union> (vars_term (ti \<rho> (length (snd \<rho>) - 1)))) - (set (X \<rho> (length (snd \<rho>) - 1)))"
proof-
 let ?k = "length (snd \<rho>)" 
 let ?f = "\<lambda>i. set (X \<rho> (Suc i)) - set (X \<rho> i)"
 from l have kl:"?k = length ((set (X \<rho> 0)) # map ?f [0..<?k - 1])" unfolding length_Cons length_map length_upt by simp
 with nth_append_length show ?thesis unfolding X_part_def append_Cons[symmetric] by (metis (no_types, lifting))
qed


lemma 
 assumes inR:"\<rho> \<in> R" 
 and ull:"ultra_rule left_linear_trs \<rho>"
 and l:"length (snd \<rho>) \<noteq> 0 "
 and sl:"length (\<sigma> :: ('f,'v) subst list) = Suc (length (snd \<rho>))"
 and st:"\<forall> i \<le> length (snd \<rho>). ((si \<rho> i) \<cdot> (\<sigma>!i), (ti \<rho> i) \<cdot> (\<sigma>!Suc i) ) \<in> (rstep UR)^*"
 and zrew:"\<forall> i < length (snd \<rho>) . \<forall> z \<in> set(Z \<rho> i) . (Var z \<cdot> (\<sigma>!i), Var z \<cdot> (\<sigma>!Suc i)) \<in> (rstep UR)^*"
 shows
 "(clhs \<rho> \<cdot> (\<sigma>!0), crhs \<rho> \<cdot> (\<sigma>!(length (snd \<rho>)))) \<in> (rstep UR)^+"
proof-
 let ?k = "length (snd \<rho>)"
 let ?xs = "X_part \<rho>"
 let ?\<sigma> = "fun_merge \<sigma> ?xs"
 from not0_implies_Suc l obtain k' where k_suc:"?k = Suc k'" by auto
 have lx:"Suc ?k = length ?xs" unfolding X_part_def using k_suc by fastforce
 from X_part[OF inR l] have p:"is_partition ?xs" by fast
 { fix t :: "('f,'v) term" fix i
   assume x:"i < Suc ?k" and y:"vars_term t \<subseteq> ?xs!i"
   from y fun_merge_part[OF p] x[unfolded lx] have "\<And> x. x \<in> (vars_term t) \<Longrightarrow> ?\<sigma> x = (\<sigma>!i) x" by blast
   from term_subst_eq this have "t \<cdot> ?\<sigma> = t \<cdot> (\<sigma>!i)" by blast
 } note subst_fact = this
 let ?l = "clhs \<rho>" let ?r = "crhs \<rho>" let ?cs = "snd \<rho>"
 from inR have rho:"((?l,?r),?cs) \<in> R " by auto
 have "vars_term ?l \<subseteq> ?xs ! 0" using l_vars_X_vars[OF inR, OF l[unfolded neq0_conv]] unfolding X_part_def by auto
 from subst_fact[OF _ this] lx have lsigma:"?l \<cdot> ?\<sigma> = ?l \<cdot> (\<sigma>!0)" by fast
 (* next, show t_i \<sigma>_{i+1} = t_i \<sigma> *)
  { fix i
    assume ik:"i < ?k"  
    from conjunct2[OF conjunct2[OF L3_8a[OF inR ull]], rule_format,OF ik] have ss':"vars_term (ti \<rho> i) \<inter> set (X \<rho> i) = {}" by auto
    have "vars_term (ti \<rho> i) \<subseteq> ?xs!Suc i"  proof(cases "Suc i < ?k")
    case True
     from ik min_less_iff_conj nth_mem[of i "take (Suc i)(snd \<rho>)", unfolded length_take nth_take[OF lessI]] 
      have "ti \<rho> i \<in> rhss (set (take (Suc i)(snd \<rho>)))" by blast
     with subsetI image_eqI[OF refl] have "vars_term (ti \<rho> i) \<subseteq> \<Union>(vars_term ` rhss (set (take (Suc i) (snd \<rho>))))" by blast
     hence ss:"vars_term (ti \<rho> i) \<subseteq> set (X \<rho> (Suc i))" 
      using conjunct1[OF Z[unfolded Z_vars_def], unfolded X_vars_def,rule_format,OF inR True] by auto
     from X_part_nth[OF inR _ True] k_suc have "?xs!Suc i = set (X \<rho> (Suc i)) - set (X \<rho> i)" by auto
     with ss ss' show "vars_term (ti \<rho> i) \<subseteq> ?xs!Suc i" by blast
    next
    case False
     with ik have i:"i = ?k - 1" and ii: "Suc i = ?k" by auto
     have xk: "?xs ! (Suc i) = vars_term (crhs \<rho>) \<union> vars_term (ti \<rho> i) - set (X \<rho> i)" 
      unfolding X_part_def i k_suc
      using nth_append_length[of "set (X \<rho> 0) # map (\<lambda>i. set (X \<rho> (Suc i)) - set (X \<rho> i)) [0..<Suc k' - 1]", unfolded length_Cons length_map length_upt]
      unfolding length_Cons length_map length_upt by force
     with ss' show "vars_term (ti \<rho> i) \<subseteq> ?xs!Suc i" unfolding ii by blast
    qed
    from subst_fact[OF _ this] ik have "(ti \<rho> i) \<cdot> (\<sigma>! (Suc i)) = (ti \<rho> i) \<cdot> ?\<sigma>" by fastforce
  } note t_fact = this 
(* attempt of more general auxiliary result for s_i and r case *)
{ fix j x
  assume y:"x \<in> set (Y \<rho> j)" 
   and lhs_or_ti:"x \<in> vars_term ?l \<or> (\<exists> k < Suc j. x \<in> vars_term (ti \<rho> k))"
   and ik:"Suc j \<le> ?k"
  from y Y_vars_mono[OF inR] Suc_le_lessD[OF ik] have y:"\<And> j'. j' \<le> j \<Longrightarrow> x \<in> set (Y \<rho> j')" by blast
  have "(?\<sigma> x, (\<sigma>! (Suc j)) x) \<in> (rstep UR)^*" 
   proof(cases "x \<in> vars_term ?l")
    case True 
     with vars_l_in_X[OF inR] k_suc have "x \<in> ?xs!0"  unfolding X_part_def nth_Cons_0 by auto
     from fun_merge_part[OF p _ this, unfolded lx[symmetric] k_suc] have  sx0:"?\<sigma> x = (\<sigma>!0) x" by auto
       from vars_l_in_X[OF inR _ True, unfolded k_suc, OF zero_less_Suc] X_vars_mono[OF inR _ le0] 
         order.strict_trans2[OF _ ik] have x:"\<And> j'. j' < Suc j \<Longrightarrow> x \<in> set (X \<rho> j')" by blast
       from x y X_Y_imp_Z[OF inR] ik k_suc have z:"\<And> j'. j' < Suc j \<Longrightarrow> x \<in> set (Z \<rho> j')" by simp
       hence xZ0:"x \<in> set (Z \<rho> 0)" by auto
       from z ik have "((\<sigma> ! 0) x,  (\<sigma> ! (Suc j)) x) \<in> (rstep UR)\<^sup>*" proof (induct j)
        case 0 from zrew[rule_format, of 0] k_suc xZ0 show ?case by simp
        next
        case (Suc k) 
         hence 1:"Suc k < ?k" and 2:"\<And> j. j < Suc k \<Longrightarrow> x \<in> set (Z \<rho> j)" by auto
         from zrew[rule_format, OF 1 Suc(2)] Suc(1) 2 1 show ?case by fastforce
        qed
       with sx0 show ?thesis by auto
      next 
      case False
       with lhs_or_ti obtain j' where ji:"j' < Suc j" and xj:"x \<in> vars_term (ti \<rho> j')" by auto
       from ji ik have 1:"Suc j' \<le> ?k" by auto
       from  L3_8a[OF inR ull] order.strict_trans2[OF ji ik] have x:"vars_term (ti \<rho> j') \<inter> set (X \<rho> j') = {}" by fast
       from nth_take[OF lessI] nth_mem[of j' "take (Suc j') (snd \<rho>)", unfolded nth_take[OF lessI] length_take min_less_iff_conj]  
        order.strict_trans2[OF ji ik] image_iff[of "ti \<rho> j'" snd] lessI[of j']
        have "ti \<rho> j' \<in>  rhss (set (take (Suc j') ?cs))" by blast
       with xj image_iff have xu:"x \<in> \<Union>(vars_term ` rhss (set (take (Suc j') ?cs)))" unfolding Union_iff by blast
       from xu conjunct1[OF Z[unfolded Z_vars_def], unfolded X_vars_def, rule_format, OF inR] False
          have xsj: "Suc j' < ?k \<Longrightarrow> x \<in> set (X \<rho> (Suc j'))" by blast
       have "x \<in> ?xs!(Suc j')"
        proof(cases "Suc j' < ?k")
         case True
          from xsj[OF True] x xj X_part_nth[OF inR _ True] k_suc show ?thesis by auto
         next
         case False
          with 1 have sk:"?k = Suc j'" by auto
          with xj x show "x \<in> ?xs!(Suc j')" unfolding X_part_last[OF inR nat.discI[OF sk], unfolded sk] by auto
        qed
        from fun_merge_part[OF p _ this] lx 1 have sxSucj:"?\<sigma> x = (\<sigma>!Suc j') x" by auto
       { fix j''
         assume 1:"Suc j' \<le> j''" and 2:"j'' < Suc j"
         with ik have ltk:"Suc j' < ?k" by force
         have "x \<in> set (X \<rho> j'')" proof(cases "Suc j' = j''", insert xsj[OF ltk],simp)
           case False with 1 have "Suc j' < j''" by auto
             from xsj[OF ltk]  X_vars_mono[OF inR _ 1] 2 ik show "x \<in> set (X \<rho> j'')" by auto
           qed 
         from X_Y_imp_Z[OF inR _ this y[OF 2[unfolded less_Suc_eq_le]]] 2 ik have "x \<in> set (Z \<rho> j'')" by auto
       } note z = this
       have "((\<sigma> ! Suc j') x,  (\<sigma> ! (Suc j)) x) \<in> (rstep UR)\<^sup>*"
       proof(cases "Suc j' = Suc j", insert ji, simp)
       case False 
        with ji have "Suc j' < Suc j" by auto
        with z ik show ?thesis proof (induct j,simp)
        case (Suc k) 
         from Suc have z:"\<And> j''. Suc j' \<le> j'' \<and> j'' < Suc k \<Longrightarrow> x \<in> set (Z \<rho> j'')" by auto
         show ?case proof(cases "Suc j'= Suc k")
         case False
          with Suc(4) Suc(3) have 0:"j' < Suc k" and 1:"Suc k < ?k" and 2:"Suc j' < Suc k" by auto
          from z Suc(1)[OF _ less_imp_le_nat[OF 1] 2] have "((\<sigma> ! Suc j') x, (\<sigma> ! Suc k) x) \<in> (rstep UR)\<^sup>*" by blast
          with zrew[rule_format, OF 1] Suc(2)[of "Suc k", OF Suc_leI[OF 0] lessI] show ?thesis by force
         next
         case True
           from Suc have 0:"Suc j' < ?k" by auto
           from True zrew[rule_format, OF 0 Suc(2)[OF order.refl Suc(4)]] show ?thesis by simp
         qed
       qed
       qed
       with sxSucj show ?thesis by auto
    qed
 } note auxiliary2 = this
(* case for s_i *)
 { fix i x
   assume ik:"i < ?k" 
   { fix x
   assume xs:"x \<in> vars_term (si \<rho> i)"
   have " (?\<sigma> x, (\<sigma>!i) x) \<in> (rstep UR)^*" proof(cases i) 
    case 0
     from xs[unfolded 0] dctrs[rule_format, OF inR ik[unfolded 0]] have  "x \<in> vars_term ?l" by fast
     with vars_l_in_X[OF inR] k_suc have "x \<in> ?xs!0"  unfolding X_part_def nth_Cons_0 by auto
     from fun_merge_part[OF p _ this, unfolded lx[symmetric] k_suc] have  "?\<sigma> x = (\<sigma>!0) x" by auto
     with 0 show ?thesis by auto
    next
    case (Suc j)
     from s_imp_Y[OF inR ik[unfolded Suc] xs[unfolded Suc]] ik have y:" x \<in> set (Y \<rho> j)" by blast
     from rev_subsetD[OF xs dctrs[rule_format, OF inR ik]] auxiliary2[OF y] Suc ik show ?thesis by fastforce
   qed
   }
   with ik substs_rsteps' have "((si \<rho> i) \<cdot> ?\<sigma>, (si \<rho> i) \<cdot> (\<sigma>!i)) \<in> (rstep UR)^*" by blast
 } note ss = this
(* case for r *)
 { fix i x
   assume xr:"x \<in> vars_term (crhs \<rho>)"
   from rev_subsetD[OF xr type3[rule_format, OF inR], unfolded Un_iff[of x]] have 
      disj:"x \<in> vars_term ?l \<or> (\<exists> j. j< ?k \<and> x \<in> vars_term (ti \<rho> j))" by fastforce
   from conjunct1[OF conjunct2[OF Z[unfolded Z_vars_def]], unfolded Y_vars_def, rule_format, OF inR] xr
     have y:" x \<in> set (Y \<rho> k')" unfolding k_suc by simp
   from auxiliary2[OF y] disj k_suc have "(?\<sigma> x, (\<sigma>!?k) x) \<in> (rstep UR)^*"  by auto
} note xr = this
 from xr substs_rsteps' have rsigma:"(?r \<cdot> ?\<sigma>, ?r \<cdot> (\<sigma>!?k)) \<in> (rstep UR)^*" by blast
(* combine to final result *)
 from ss st[rule_format] t_fact have "\<And> i. i  < ?k \<Longrightarrow> ((si \<rho> i) \<cdot> ?\<sigma>,(ti \<rho> i) \<cdot> ?\<sigma>) \<in> (rstep UR)^*" 
  using rtrancl_trans[of _ _ "rstep UR"] nat_less_le by metis
 with UR_simulation_main[OF rho order.refl, of ?\<sigma>] have "(?l \<cdot> ?\<sigma>, rhs_n \<rho> ?k \<cdot> ?\<sigma>) \<in> (rstep UR)\<^sup>+" by simp
 from this[unfolded lsigma k_suc] rsigma show ?thesis 
  unfolding rhs_n.simps[of "fst \<rho>" "snd \<rho>" ?k, unfolded k_suc prod.collapse] by force
qed

definition non_LV :: bool where "non_LV \<equiv> \<forall> (l,r) \<in> UR. \<not>is_Var l"


lemma unnamed1:
 assumes step:"(s,t) \<in> rstep UR"
 and rhoR:"\<rho> \<in> R"
 and sroot:"s = Fun Um sa" 
 and u:"U \<rho> m = (More Um Nil Hole (map Var (Z \<rho> m)))" "Um \<notin> funs_ctrs R" "m < length (snd \<rho>)"
 and troot:"(t = Fun f ta  \<and> f \<noteq> Um) \<or> (t = Var z)"
 and nlv:non_LV 
 shows "\<exists> \<sigma>. (s,t) \<in> rstep_r_p_s UR (lhs_n \<rho> (Suc m), rhs_n \<rho> (Suc m)) \<epsilon> \<sigma>"
proof-
 from step[unfolded rstep_iff_rstep_r_p_s] obtain l r p \<sigma> where 
   step:"(l,r) \<in> UR \<and> s |_ p = fst (l, r) \<cdot> \<sigma> \<and> t = (ctxt_of_pos_term p s)\<langle>r \<cdot> \<sigma>\<rangle>" 
   unfolding rstep_r_p_s_def' mem_Collect_eq split by force
 from conjunct2[OF this, unfolded sroot] troot have
   x:"Fun Um sa |_ p = fst (l, r) \<cdot> \<sigma>" 
   " (Var z = (ctxt_of_pos_term p (Fun Um sa))\<langle>r \<cdot> \<sigma>\<rangle>) \<or> (Fun f ta = (ctxt_of_pos_term p (Fun Um sa))\<langle>r \<cdot> \<sigma>\<rangle> \<and> f \<noteq> Um)"
   (is "?var \<or> ?fun") by auto
 have pe:"p = \<epsilon>" proof(cases p, simp_all)
   fix j q
   assume p:"p = j <# q"
   have "\<not> ?var" unfolding p ctxt_of_pos_term.simps(2)  ctxt_apply.simps(2) by fast
   from p Meson.make_pos_rule'[OF x(2) this, unfolded this ctxt_of_pos_term.simps(2) ctxt_apply.simps(2)] 
     show False by fastforce
   qed
 with nlv conjunct1[OF step] x obtain ls where lUm:"l = Fun Um ls" unfolding non_LV_def by fastforce
 from U_cond[rule_format, OF rhoR u(3)] u(1) have fresh:"Um \<notin> funs_ctrs R \<and>
   \<not> (\<exists>\<rho>' n' b c a. \<rho>' \<in> R \<and> n' < length (snd \<rho>') \<and> (\<rho> \<noteq> \<rho>' \<or> m \<noteq> n') \<and> U \<rho>' n' = More Um b c a)" by simp
 have "\<And> l' r'. ((l',r') \<in> UR \<Longrightarrow> \<exists> us. l' = Fun Um us \<Longrightarrow> (l',r') = (lhs_n \<rho> (Suc m), rhs_n \<rho> (Suc m)))" proof-
   fix l' r'
   assume lr':"(l',r') \<in> UR" and "\<exists> us. l' = Fun Um us"
   then obtain us where l':"l' = Fun Um us" by auto
   from lr'[unfolded UR_def image_iff] have "\<exists> \<rho>' \<in> R. (l',r') \<in> rules \<rho>'" by blast
   then obtain \<rho>' where rho':"\<rho>' \<in> R""(l',r') \<in> rules \<rho>'" by blast
   from this(2)[unfolded rules_def image_iff] mem_Collect_eq lhs_n_def have
     "\<exists> k. k \<le> length (snd \<rho>') \<and> l' = lhs_n \<rho>' k \<and> r' = rhs_n \<rho>' k" by blast
  then obtain k where k:"k \<le> length (snd \<rho>')" "l' = lhs_n \<rho>' k" "r' = rhs_n \<rho>' k" by auto
  from k(2) lhs_n.simps prod.collapse have x:"l' = clhs \<rho>' \<or> (\<exists> k'. k = (Suc k') \<and> l' = (U \<rho>' k')\<langle>snd ((snd \<rho>') ! k')\<rangle>)"
    by (metis not0_implies_Suc)
  show "(l',r') = (lhs_n \<rho> (Suc m), rhs_n \<rho> (Suc m))" proof(cases "l' = clhs \<rho>'")
   case True
    from this[unfolded l'] rho'(1) have "Um \<in> funs_rule (fst \<rho>')" unfolding funs_rule_def by simp
    with conjunct1[OF fresh] rho'(1) have False unfolding l' using image_iff by blast
    thus ?thesis by auto
   next
   case False
    with x obtain k' where k':"k = (Suc k')" "l' = (U \<rho>' k')\<langle>snd ((snd \<rho>') ! k')\<rangle>" by auto
    from k'(2) ctxt_apply.simps l' have z:"\<exists> a b c. U \<rho>' k' = More Um b c a" proof(cases "U \<rho>' k'", auto)
     assume a:"ti \<rho>' k' = Fun Um us"
     from a have y:"Um \<in> funs_rule ((snd \<rho>') ! k')" unfolding funs_rule_def Un_iff by simp
     from k(1)[unfolded k'(1)] have "k' < length (snd \<rho>')" by fastforce
     from y nth_mem[OF this] have  "Um \<in> funs_trs (set (snd \<rho>'))" unfolding funs_trs_def UN_iff by fast
     with conjunct1[OF fresh] rho'(1) show False by blast
    qed
    from k(1) k'(1) have "k' < length (snd \<rho>')" by auto
    with conjunct2[OF fresh] rho'(1) z have "\<rho> = \<rho>' \<and> m = k'" by blast
    with k' k show ?thesis by fast
   qed
  qed
  with lUm conjunct1[OF step] have "(l, r) = (lhs_n \<rho> (Suc m), rhs_n \<rho> (Suc m))" by force
  with step[unfolded pe] rhoR ctxt_supt_id[OF empty_pos_in_poss] show ?thesis 
    unfolding rstep_r_p_s_def mem_Collect_eq split by fastforce
qed

lemma unnamed2:
 assumes step:"(s,t) \<in> rstep UR"
 and rhoR:"\<rho> \<in> R"
 and sroot:"s = Fun Um sa" 
 and u:"U \<rho> m = (More Um Nil Hole (map Var (Z \<rho> m)))" "Um \<notin> funs_ctrs R" "m < length (snd \<rho>)"
 and troot:"t = Var z"
 and nlv:non_LV
 shows "(s,t) \<in> rstep {(lhs_n \<rho> (Suc m), rhs_n \<rho> (Suc m))}"
 oops

lemma Um_funs_R:
 assumes um:"U \<rho> m = (More Um Nil Hole (map Var (Z \<rho> m)))" and len:"m < length (snd \<rho>)"
 and rhoR:"\<rho> \<in> R"
 shows "Um \<notin> funs_ctrs R"
 using U_cond[rule_format, OF rhoR len] um by auto
   
lemma no_vars_in_seq:
 assumes seq:"\<forall>i<length us - 1. (us ! i, us ! Suc i) \<in> rstep UR"
 and nonlv:non_LV
 shows "\<And> i. i < (length us) - 1 \<Longrightarrow> is_Fun (us ! i)"
proof
 fix i
 assume i:"i < (length us) - 1" and "is_Var (us ! i)"
 then obtain x where v:"us ! i = Var x" by auto
 from seq[rule_format, OF i, unfolded rstep_iff_rstep_r_p_s rstep_r_p_s_def'] obtain l r p \<sigma> where
     m:"(l, r) \<in> UR \<and> p \<in> poss (us ! i) \<and> (us ! i) |_ p = l \<cdot> \<sigma>" by force 
 from conjunct2[OF this, unfolded v poss.simps(1) singleton_iff] have "is_Var l"
     by (metis is_VarI subst_apply_eq_Var subt_at.simps(1))
 with nonlv[unfolded non_LV_def] conjunct1[OF m] show False by fast
qed

lemma about_roots1:
 assumes seq:"(\<forall>i<(length us) - 1. (us ! i, us ! Suc i) \<in> rstep UR)"
 and froot:"root (us ! ((length us) - 1)) = Some (f,a) \<and> f \<in> funs_ctrs R"
 and rhoR:"\<rho> \<in> R"
 and uroot:"root (us ! 0) = Some (Um,a')" "U \<rho> m = (More Um Nil Hole (map Var (Z \<rho> m)))" (*"Um \<notin> funs_ctrs R"*) "m < length (snd \<rho>)"
 and nonlv:"non_LV"
 shows "(\<forall>i. m < i \<and> i \<le> length (snd \<rho>) \<longrightarrow>
  (\<exists> j < (length us) - 1. \<exists> \<sigma> . (us ! j, us ! Suc j) \<in> (rstep_r_p_s UR (lhs_n \<rho> i, rhs_n \<rho> i) \<epsilon> \<sigma> )))"
proof-
 from seq froot uroot show ?thesis proof(induct "length us" arbitrary: us Um a' m)
  case 0 
   with Um_funs_R[OF 0(5) 0(6) rhoR] show ?case by auto
  next
  case (Suc n)
   let ?us' = "tl us"
   let ?step = "\<lambda> ts j i \<sigma> . (ts ! j, ts ! Suc j) \<in> (rstep_r_p_s UR (lhs_n \<rho> i, rhs_n \<rho> i) \<epsilon> \<sigma> )"
   let ?ur_step = "?step us" and ?ur_step' = "?step ?us'"
   let ?k = "length (snd \<rho>)"
   let ?all_steps = "\<lambda> m b us. \<forall>i. m < i \<and> i \<le> ?k \<longrightarrow> (\<exists>j < b. \<exists>\<sigma>. ?step us j i \<sigma>)" 
   from Suc(2) have 1:"length ?us' = n" by auto
   from Um_funs_R[OF Suc(6) Suc(7) rhoR] have UmR:"Um \<notin> funs_ctrs R" by auto
   with conjunct2[OF Suc(4)] have "f \<noteq> Um" by auto
   with conjunct1[OF Suc(4)] Suc(5) option.inject[of "(f,a)" "(Um,a')", unfolded old.prod.inject] Suc(2)
    have lp:"Suc n > 1" unfolding Suc(2)[symmetric] by fastforce
   note nth_tl' = nth_tl[of _ us, unfolded 1]
   have 3:"(\<forall>i < n - 1. (?us' ! i, ?us' ! Suc i) \<in> rstep UR)"
     proof(rule, rule)
      fix i
      assume a:"i < n - 1"
      hence "i < n" "Suc i < n" by auto 
      from nth_tl'[OF this(1)]  nth_tl'[OF this(2)]  
       Suc(3)[rule_format] Suc(2) a show "(?us' ! i, ?us' ! Suc i) \<in> rstep UR" by auto
     qed
    from lp 1 have npos:"0 < n" by auto
    from nth_tl[of "n - 1" us, unfolded 1] npos have "(us ! n) = (?us' ! (n - 1))" 
     unfolding 1 Suc(2)[symmetric] by force
    with Suc(4) 1 have 4:"root (?us' ! (n - 1)) = Some (f,a) \<and> f \<in> funs_ctrs R" by auto
   show ?case proof(cases "\<exists> a''. root (us ! 1) = Some(Um,a'')")
   case True
    then obtain a'' where True':"root (us ! 1) = Some(Um,a'')" by auto
    from True' lp nth_tl[of 0 us, unfolded 1, OF npos] have 5:"root (tl us ! 0) = Some (Um, a'')" by force
    (* case 1: root symbol remains the same *)
    note ih = Suc(1)[OF 1[symmetric], unfolded 1, OF 3 4 5 Suc(6) Suc(7), unfolded length_tl]
    show ?thesis proof(rule,rule)
     fix i 
     assume a:"m < i \<and> i \<le> ?k"
     from ih[rule_format, OF a] obtain j \<sigma> where b:"j < n - 1" "?ur_step' j i \<sigma>" by force
     from b(1) have l:"Suc j < n" by force
     from nth_tl'[OF Suc_lessD[OF this]] nth_tl'[OF this] b(2) have "?ur_step (Suc j) i \<sigma>" by auto
     with l[unfolded length_tl] Suc(2) show "\<exists>j<length us - 1. \<exists>\<sigma>.  ?ur_step j i \<sigma>" by force
    qed
   next
   (* case 2: root symbol changes *)
   case False note change=this
    from Suc(3)[rule_format, of 0] Suc(2) lp have step:"(us ! 0, us ! Suc 0) \<in> rstep UR" by force
    with rstep_iff_rstep_r_p_s obtain l r p \<sigma> where "(us ! 0, us ! Suc 0) \<in> rstep_r_p_s UR (l, r) p \<sigma>" by metis
    from Suc(5) Some_root_imp_Fun obtain ss where f0:"us ! 0 = Fun Um ss" by fast
    let ?l = "clhs \<rho>" and ?r = "crhs \<rho>" and ?cs = "snd \<rho>"
    from prod.collapse have rho:"\<rho> = ((?l,?r),?cs)" by simp
    with Suc(2) lp have l0:"0 < length us - 1" by auto
    have "\<exists> g ss. us ! 1 = Fun g ss \<and> g \<noteq> Um" proof(cases "Suc n = 2")
       case True
        from True Suc(2) Suc(4) have x:"root (us ! 1) = Some (f, a) \<and> f \<in> funs_ctrs R" by force
        with Some_root_imp_Fun obtain ss where "us ! 1 = Fun f ss \<and> f \<in> funs_ctrs R" by metis
        with UmR show ?thesis by blast
       next
       case False
        with Suc(2) l0 have "1 < length us - 1" by linarith
        from no_vars_in_seq[OF Suc(3) nonlv this, unfolded is_Fun_Fun_conv] change show ?thesis by auto
       qed
     then obtain g ss where gUm:"us ! Suc 0 = Fun g ss \<and> g \<noteq> Um" by auto
     with unnamed1[OF step rhoR f0 Suc(6) UmR Suc(7) _ nonlv] obtain \<sigma> where step:"?ur_step 0 (Suc m) \<sigma>" by auto
     hence rhs:"us ! 1 = rhs_n \<rho> (Suc m) \<cdot> \<sigma>" unfolding rstep_r_p_s_def by auto
     from gUm nth_tl[unfolded length_tl, OF l0] obtain a' where 5:"root (?us' ! 0) = Some (g, a')" by simp
     show ?thesis proof (cases "Suc m < length (snd \<rho>)")
      case True
       with rhs rhs_n.simps[of "fst \<rho>" "snd \<rho>" "Suc m"] have us1:"us ! 1 = (U (fst \<rho>, snd \<rho>) (Suc m))\<langle>si \<rho> (Suc m)\<rangle> \<cdot> \<sigma>" by force
       from U_cond[rule_format, OF rhoR True] obtain Um' where 
        x:"U \<rho> (Suc m) = More Um' [] \<box> (map Var (Z \<rho> (Suc m))) \<and> Um' \<notin> funs_ctrs R" by auto
       with us1[unfolded pair_collapse] obtain a' where Um':"root (us ! 1) = Some (Um', a')" by auto
       from nth_tl[unfolded length_tl, OF l0] Um' have 5:"root (?us' ! 0) = Some (Um', a')" by simp
       from Suc(1)[OF 1[symmetric], unfolded 1, OF 3 4 5 _ True] x have y:"?all_steps (Suc m) (n - 1) ?us'" by auto
       show "?all_steps m (length us - 1) us"
       proof(rule, rule impI)
        fix i
        assume a:"m < i \<and> i \<le> ?k"
        show "\<exists>j<length us - 1. \<exists>\<sigma>.?ur_step j i \<sigma>" proof(cases "i = Suc m")
         case True
          from step l0 show ?thesis unfolding True by force
         next
         case False
          with a y[rule_format] 1 obtain j \<sigma> where b:"j < length ?us' - 1 \<and> ?ur_step' j i \<sigma>" by fastforce
          from b length_tl have c:"j < length us - 1" "Suc j < length us - 1" by auto
          from conjunct2[OF b] nth_tl[unfolded length_tl, OF c(1)] nth_tl[unfolded length_tl, OF c(2)] have
           "?ur_step (Suc j) i \<sigma>" by auto
          with c(2) show "\<exists>j<length us - 1. \<exists>\<sigma>.?ur_step j i \<sigma>" by blast
         qed
       qed
      next
      case False
       with Suc(7) have len:"Suc m = length (snd \<rho>)" by auto
       show ?thesis proof(rule,rule)
        fix i 
        assume a:"m < i \<and> i \<le> length (snd \<rho>)"
        with len have ism:"i = Suc m" by auto
        from len rhs[unfolded rhs_n.simps[of "(?l,?r)" ?cs, unfolded rho[symmetric]]] have "us ! 1 = crhs \<rho> \<cdot> \<sigma>" by auto
        from step[unfolded ism[symmetric]] Suc(2) npos show " \<exists>j<length us - 1. \<exists>\<sigma>. ?ur_step j i \<sigma>" by force
       qed
     qed
    qed
   qed
qed

end

end

